/* This file is automatically generated by buildconfig */

#ifndef OVMETHODS_H
#define OVMETHODS_H 1

#include "ovinterface.h"

#define NUM_OV_METHODS 3

extern OV_METHOD ov_methods[NUM_OV_METHODS];

#endif /* OVMETHODS_H */
