/* This file is automatically generated by buildconfig. */

#include "ovinterface.h"
#include "buffindexed/buffindexed.h"
#include "ovdb/ovdb.h"
#include "tradindexed/tradindexed.h"

OV_METHOD ov_methods[3] = {
{
    "buffindexed",
    buffindexed_open,
    buffindexed_groupstats,
    buffindexed_groupadd,
    buffindexed_groupdel,
    buffindexed_add,
    buffindexed_cancel,
    buffindexed_opensearch,
    buffindexed_search,
    buffindexed_closesearch,
    buffindexed_getartinfo,
    buffindexed_expiregroup,
    buffindexed_ctl,
    buffindexed_close
},
{
    "ovdb",
    ovdb_open,
    ovdb_groupstats,
    ovdb_groupadd,
    ovdb_groupdel,
    ovdb_add,
    ovdb_cancel,
    ovdb_opensearch,
    ovdb_search,
    ovdb_closesearch,
    ovdb_getartinfo,
    ovdb_expiregroup,
    ovdb_ctl,
    ovdb_close
},
{
    "tradindexed",
    tradindexed_open,
    tradindexed_groupstats,
    tradindexed_groupadd,
    tradindexed_groupdel,
    tradindexed_add,
    tradindexed_cancel,
    tradindexed_opensearch,
    tradindexed_search,
    tradindexed_closesearch,
    tradindexed_getartinfo,
    tradindexed_expiregroup,
    tradindexed_ctl,
    tradindexed_close
}
};

